from cfelpyutils.crystfel_utils import load_crystfel_geometry
from concurrent.futures import ThreadPoolExecutor
from itertools import product
from matplotlib.axes import Axes
import numpy as np
import pytest

from extra_geom import AGIPD_1MGeometry, agipd_asic_seams


def test_snap_assemble_data():

    def check_result(img, centre):
        assert img.shape == (1256, 1092)
        assert tuple(centre) == (631, 550)
        assert np.isnan(img[0, 0])
        assert img[50, 50] == 0

    geom = AGIPD_1MGeometry.from_quad_positions(
        quad_pos=[(-525, 625), (-550, -10), (520, -160), (542.5, 475)]
    )

    stacked_data = np.zeros((16, 512, 128))
    img, centre = geom.position_modules_fast(stacked_data)
    check_result(img, centre)

    # test unsafe cast with output array
    stacked_data = np.zeros((16, 512, 128), dtype=np.float64)
    out = geom.output_array_for_position_fast(dtype=np.float32)
    with pytest.raises(TypeError):
        img, centre = geom.position_modules_fast(stacked_data, out=out)

    # test safe cast with output array
    stacked_data = np.zeros((16, 512, 128), dtype=np.uint16)
    out = geom.output_array_for_position_fast(dtype=np.float32)
    img, centre = geom.position_modules_fast(stacked_data, out=out)
    check_result(img, centre)
    check_result(out, centre)
    assert img.dtype == out.dtype == np.float32

    # Assemble in parallel
    stacked_data = np.zeros((16, 512, 128))
    with ThreadPoolExecutor(max_workers=2) as tpool:
        img, centre = geom.position_modules_fast(stacked_data, threadpool=tpool)
        check_result(img, centre)


def test_assemble_symmetric():
    geom = AGIPD_1MGeometry.from_quad_positions(
        quad_pos=[(-525, 625), (-550, -10), (520, -160), (542.5, 475)]
    )
    print("2 centre", geom._snapped().centre * 2)

    stacked_data = np.zeros((16, 512, 128))
    img = geom.position_modules_symmetric(stacked_data)

    assert img.shape == (1262, 1100)
    assert np.isnan(img[0, 0])
    assert np.isnan(img[img.shape[0] // 2, img.shape[1] // 2])
    assert img[50, 50] == 0

    # Smoketest assembling into suitable output array
    geom.position_modules_symmetric(stacked_data, out=img)

    with pytest.raises(ValueError):
        # Output array not big enough
        geom.position_modules_symmetric(stacked_data, out=img[:-1, :-1])


def test_write_read_crystfel_file(tmpdir):
    geom = AGIPD_1MGeometry.from_quad_positions(
        quad_pos=[(-525, 625), (-550, -10), (520, -160), (542.5, 475)]
    )
    path = str(tmpdir / 'test.geom')
    geom.write_crystfel_geom(filename=path, photon_energy=9000,
                             adu_per_ev=0.0075, clen=0.2)

    loaded = AGIPD_1MGeometry.from_crystfel_geom(path)
    np.testing.assert_allclose(
        loaded.modules[0][0].corner_pos, geom.modules[0][0].corner_pos
    )
    np.testing.assert_allclose(loaded.modules[0][0].fs_vec, geom.modules[0][0].fs_vec)

    # Load the geometry file with cfelpyutils and test the rigid groups
    geom_dict = load_crystfel_geometry(path)
    quad_gr0 = [  # 1st quadrant: p0a0 ... p3a7
        'p{}a{}'.format(p, a) for p, a in product(range(4), range(8))
    ]
    assert geom_dict['rigid_groups']['p0'] == quad_gr0[:8]
    assert geom_dict['rigid_groups']['p3'] == quad_gr0[-8:]
    assert geom_dict['rigid_groups']['q0'] == quad_gr0
    assert geom_dict['panels']['p0a0']['res'] == 5000  # 5000 pixels/metre
    p3a7 = geom_dict['panels']['p3a7']
    assert p3a7['min_ss'] == 448
    assert p3a7['max_ss'] == 511
    assert p3a7['min_fs'] == 0
    assert p3a7['max_fs'] == 127


def test_write_read_crystfel_file_2d(tmpdir):
    geom = AGIPD_1MGeometry.from_quad_positions(
        quad_pos=[(-525, 625), (-550, -10), (520, -160), (542.5, 475)]
    )
    path = str(tmpdir / 'test.geom')
    geom.write_crystfel_geom(filename=path, dims=('frame', 'ss', 'fs'),
                             adu_per_ev=0.0075, clen=0.2)

    loaded = AGIPD_1MGeometry.from_crystfel_geom(path)
    np.testing.assert_allclose(
        loaded.modules[0][0].corner_pos, geom.modules[0][0].corner_pos
    )
    np.testing.assert_allclose(loaded.modules[0][0].fs_vec, geom.modules[0][0].fs_vec)

    # Load the geometry file with cfelpyutils and check some values
    geom_dict = load_crystfel_geometry(path)

    p3a7 = geom_dict['panels']['p3a7']
    assert p3a7['dim_structure'] == ['%', 'ss', 'fs']
    assert p3a7['min_ss'] == (3 * 512) + 448
    assert p3a7['max_ss'] == (3 * 512) + 511
    assert p3a7['min_fs'] == 0
    assert p3a7['max_fs'] == 127


def test_quad_positions():
    quad_pos = [(-525, 625), (-550, -10), (520, -160), (542.5, 475)]
    geom = AGIPD_1MGeometry.from_quad_positions(quad_pos)

    np.testing.assert_allclose(geom.quad_positions(), quad_pos)


def test_inspect():
    geom = AGIPD_1MGeometry.from_quad_positions(
        quad_pos=[(-525, 625), (-550, -10), (520, -160), (542.5, 475)]
    )
    # Smoketest
    ax = geom.inspect()
    assert isinstance(ax, Axes)


def test_compare():
    geom1 = AGIPD_1MGeometry.from_quad_positions(
        quad_pos=[(-525, 625), (-550, -10), (520, -160), (542.5, 475)]
    )
    geom2 = AGIPD_1MGeometry.from_quad_positions(
        quad_pos=[(-527, 625), (-548, -10), (520, -162), (542.5, 473)]
    )
    # Smoketest
    ax = geom1.compare(geom2)
    assert isinstance(ax, Axes)


def test_to_distortion_array():
    geom = AGIPD_1MGeometry.from_quad_positions(
        quad_pos=[(-525, 625), (-550, -10), (520, -160), (542.5, 475)]
    )
    # Smoketest
    distortion = geom.to_distortion_array()
    assert isinstance(distortion, np.ndarray)
    assert distortion.shape == (8192, 128, 4, 3)

    # Coordinates in m, origin at corner; max x & y should be ~ 25cm
    assert 0.20 < distortion[..., 1].max() < 0.30
    assert 0.20 < distortion[..., 2].max() < 0.30
    assert 0.0 <= distortion[..., 1].min() < 0.01
    assert 0.0 <= distortion[..., 2].min() < 0.01

def test_get_pixel_positions():
    geom = AGIPD_1MGeometry.from_quad_positions(
        quad_pos=[(-525, 625), (-550, -10), (520, -160), (542.5, 475)]
    )

    pixelpos = geom.get_pixel_positions()
    assert pixelpos.shape == (16, 512, 128, 3)
    px = pixelpos[..., 0]
    py = pixelpos[..., 1]

    assert -0.12 < px.min() < -0.1
    assert  0.12 > px.max() > 0.1
    assert -0.14 < py.min() < -0.12
    assert  0.14 > py.max() >  0.12

def test_data_coords_to_positions():
    geom = AGIPD_1MGeometry.from_quad_positions(
        quad_pos=[(-525, 625), (-550, -10), (520, -160), (542.5, 475)]
    )

    module_no = np.zeros(16, dtype=np.int16)
    slow_scan = np.linspace(0, 500, num=16, dtype=np.float32)
    fast_scan = np.zeros(16, dtype=np.float32)

    res = geom.data_coords_to_positions(module_no, slow_scan, fast_scan)

    assert res.shape == (16, 3)

    resx, resy, resz = res.T

    np.testing.assert_allclose(resz, 0)
    np.testing.assert_allclose(resy, 625 * geom.pixel_size)

    assert (np.diff(resx) > 0).all()   # Monotonically increasing
    np.testing.assert_allclose(resx[0], -525 * geom.pixel_size)
    assert -0.01 < resx[-1] < 0.01


def test_asic_seams():
    arr = agipd_asic_seams()
    assert arr.shape == (512, 128)
    assert not arr[30, 0]
    assert arr[63, 0]



import pytest
import numpy as np
from matplotlib.axes import Axes

from .utils import assert_geom_close
from extra_geom import PNCCDGeometry

def test_positions():
    # Positional constants
    gap_pixels = 50
    offset_pixels = 10

    # Calculate parameters for the relative/absolute methods
    top_offset = (-offset_pixels * PNCCDGeometry.pixel_size, 0, 0)
    bottom_offset = (offset_pixels * PNCCDGeometry.pixel_size, 0, 0)
    absolute_pos = np.array(
        [(-PNCCDGeometry.frag_fs_pixels // 2 - offset_pixels,
          PNCCDGeometry.frag_ss_pixels + gap_pixels // 2,
          0),
         (-PNCCDGeometry.frag_fs_pixels // 2 + offset_pixels,
          -gap_pixels // 2,
          0)]
    ) * PNCCDGeometry.pixel_size

    # Create the geometries
    absolute_geom = PNCCDGeometry.from_absolute_positions(top=absolute_pos[0], bottom=absolute_pos[1])
    relative_geom = PNCCDGeometry.from_relative_positions(gap=gap_pixels * PNCCDGeometry.pixel_size,
                                                          top_offset=top_offset,
                                                          bottom_offset=bottom_offset)

    # They should be the same
    assert_geom_close(absolute_geom, relative_geom)

# No re-shaping is required for the first shape, the class should automatically
# reshape the data with the second shape.
@pytest.mark.parametrize("shape", [PNCCDGeometry.expected_data_shape,
                                   (PNCCDGeometry.frag_ss_pixels * PNCCDGeometry.n_modules, PNCCDGeometry.frag_fs_pixels)])
def test_snap_assemble_data(shape):
    gap_pixels = 100
    geom = PNCCDGeometry.from_relative_positions(gap=gap_pixels * PNCCDGeometry.pixel_size)

    stacked_data = np.zeros(shape)
    img, centre = geom.position_modules_fast(stacked_data)
    expected_img_shape = (PNCCDGeometry.frag_ss_pixels * PNCCDGeometry.n_modules + gap_pixels,
                          PNCCDGeometry.frag_fs_pixels)

    assert img.shape == expected_img_shape
    assert tuple(centre) == (expected_img_shape[0] // 2, expected_img_shape[1] // 2)
    # Everywhere in the gap should be NaN
    assert np.isnan(img[img.shape[0] // 2, 0])
    assert img[50, 50] == 0

def test_inspect():
    geom = PNCCDGeometry.from_relative_positions()

    # Smoketest
    ax = geom.inspect()
    assert isinstance(ax, Axes)



from cfelpyutils.crystfel_utils import load_crystfel_geometry
from concurrent.futures import ThreadPoolExecutor
from extra_data.stacking import stack_detector_data
from itertools import product
from matplotlib.axes import Axes
import numpy as np
import pytest
import xarray as xr

from extra_geom import AGIPD_500K2GGeometry


def test_snap_assemble_data():

    def check_result(img, centre):
        assert img.shape == (602, 1068)
        assert tuple(centre) == (-100, -100)
        assert np.isnan(img[0, 535])
        assert img[50, 50] == 0

    geom = AGIPD_500K2GGeometry.from_origin((-100, -100))

    stacked_data = np.zeros((8, 512, 128))
    img, centre = geom.position_modules_fast(stacked_data)
    check_result(img, centre)

    # test unsafe cast with output array
    stacked_data = np.zeros((8, 512, 128), dtype=np.float64)
    out = geom.output_array_for_position_fast(dtype=np.float32)
    with pytest.raises(TypeError):
        img, centre = geom.position_modules_fast(stacked_data, out=out)

    # test safe cast with output array
    stacked_data = np.zeros((8, 512, 128), dtype=np.uint16)
    out = geom.output_array_for_position_fast(dtype=np.float32)
    img, centre = geom.position_modules_fast(stacked_data, out=out)
    check_result(img, centre)
    check_result(out, centre)
    assert img.dtype == out.dtype == np.float32

    # Assemble in parallel
    stacked_data = np.zeros((8, 512, 128))
    with ThreadPoolExecutor(max_workers=2) as tpool:
        img, centre = geom.position_modules_fast(stacked_data, threadpool=tpool)
        check_result(img, centre)

    # Assemble from xarray
    xr_data = xr.DataArray(
        np.zeros((7, 512, 128)),
        dims=('module', 'slow_scan', 'fast_scan'),
        coords={'module': list(range(0, 2)) + list(range(3, 8))}
    )
    img, centre = geom.position_modules_fast(xr_data)
    check_result(img, centre)

    with pytest.raises(ValueError):  # Bad module no (8)
        geom.position_modules_fast(xr_data.assign_coords(module=range(2, 9)))
    with pytest.raises(ValueError):  # No dimension named 'module'
        geom.position_modules_fast(xr_data.rename({'module': 'foo'}))

    # Assemble from an extra_data StackView
    mod_arr = np.zeros((3, 512, 128))
    d = {f'SPB_DET_AGIPD1M-1/DET/{i}CH0': {'image.data': mod_arr} for i in range(8)}
    stack = stack_detector_data(d, 'image.data', modules=8, real_array=False)
    imgs, centre = geom.position_modules_fast(stack)
    assert imgs.shape == (3, 602, 1068)
    check_result(imgs[0], centre)

def test_write_read_crystfel_file(tmpdir):
    geom = AGIPD_500K2GGeometry.from_origin()
    path = str(tmpdir / 'test.geom')
    geom.write_crystfel_geom(filename=path, photon_energy=9000,
                             adu_per_ev=0.0075, clen=0.2)

    loaded = AGIPD_500K2GGeometry.from_crystfel_geom(path)
    np.testing.assert_allclose(
        loaded.modules[0][0].corner_pos, geom.modules[0][0].corner_pos
    )
    np.testing.assert_allclose(loaded.modules[0][0].fs_vec, geom.modules[0][0].fs_vec)

    # Load the geometry file with cfelpyutils and test the rigid groups
    geom_dict = load_crystfel_geometry(path)
    quad_gr0 = [  # quadrant: p0a0 ... p7a7
        'p{}a{}'.format(p, a) for p, a in product(range(8), range(8))
    ]
    assert geom_dict['rigid_groups']['p0'] == quad_gr0[:8]
    assert geom_dict['rigid_groups']['p7'] == quad_gr0[-8:]
    assert geom_dict['rigid_groups']['q0'] == quad_gr0
    assert geom_dict['panels']['p0a0']['res'] == 5000  # 5000 pixels/metre
    p3a7 = geom_dict['panels']['p3a7']
    assert p3a7['min_ss'] == 448
    assert p3a7['max_ss'] == 511
    assert p3a7['min_fs'] == 0
    assert p3a7['max_fs'] == 127


def test_write_read_crystfel_file_2d(tmpdir):
    geom = AGIPD_500K2GGeometry.from_origin()
    path = str(tmpdir / 'test.geom')
    geom.write_crystfel_geom(filename=path, dims=('frame', 'ss', 'fs'),
                             adu_per_ev=0.0075, clen=0.2)

    loaded = AGIPD_500K2GGeometry.from_crystfel_geom(path)
    np.testing.assert_allclose(
        loaded.modules[0][0].corner_pos, geom.modules[0][0].corner_pos
    )
    np.testing.assert_allclose(loaded.modules[0][0].fs_vec, geom.modules[0][0].fs_vec)

    # Load the geometry file with cfelpyutils and check some values
    geom_dict = load_crystfel_geometry(path)

    p3a7 = geom_dict['panels']['p3a7']
    assert p3a7['dim_structure'] == ['%', 'ss', 'fs']
    assert p3a7['min_ss'] == (3 * 512) + 448
    assert p3a7['max_ss'] == (3 * 512) + 511
    assert p3a7['min_fs'] == 0
    assert p3a7['max_fs'] == 127


def test_inspect():
    geom = AGIPD_500K2GGeometry.from_origin()
    # Smoketest
    ax = geom.inspect()
    assert isinstance(ax, Axes)


def test_compare():
    geom1 = AGIPD_500K2GGeometry.from_origin((5, 10))
    geom2 = AGIPD_500K2GGeometry.from_origin((7, -5))
    # Smoketest
    ax = geom1.compare(geom2)
    assert isinstance(ax, Axes)


def test_to_distortion_array():
    geom = AGIPD_500K2GGeometry.from_origin()
    # Smoketest
    distortion = geom.to_distortion_array()
    assert isinstance(distortion, np.ndarray)
    assert distortion.shape == (4096, 128, 4, 3)

    # Coordinates in m, origin at corner; max x ~ 21cm & y ~ 12cm
    assert 0.12 < distortion[..., 1].max() < 0.121
    assert 0.21 < distortion[..., 2].max() < 0.22
    assert 0.0 <= distortion[..., 1].min() < 0.01
    assert 0.0 <= distortion[..., 2].min() < 0.01


def test_get_pixel_positions():
    geom = AGIPD_500K2GGeometry.from_origin()

    pixelpos = geom.get_pixel_positions()
    assert pixelpos.shape == (8, 512, 128, 3)
    px = pixelpos[..., 0]
    py = pixelpos[..., 1]

    assert 0. < px.min() < .0002
    assert 0.218 > px.max() > 0.21
    assert 0.13 > py.max() > 0.
    assert  0. < py.min() < 0.0002

def test_data_coords_to_positions():
    geom = AGIPD_500K2GGeometry.from_origin((0, -100))

    module_no = np.full(8, fill_value=6, dtype=np.int16)
    slow_scan = np.linspace(0, 511, num=8, dtype=np.float32)
    fast_scan = np.zeros(8, dtype=np.float32)

    res = geom.data_coords_to_positions(module_no, slow_scan, fast_scan)

    assert res.shape == (8, 3)
    print(res)

    resx, resy, resz = res.T

    np.testing.assert_allclose(resz, 0)
    np.testing.assert_allclose(resy, 100 * geom.pixel_size)

    assert (np.diff(resx) < 0).all()   # Monotonically decreasing
    np.testing.assert_allclose(resx[0], 526 * geom.pixel_size)
    assert -0.01 < resx[-1] < 0.01




from cfelpyutils.crystfel_utils import load_crystfel_geometry
import numpy as np

from extra_geom import JUNGFRAUGeometry
from .utils import assert_geom_close

def jf4m_geometry():
    x_start, y_start = 1125, 1078
    mod_width = (256 * 4) + (2 * 3)
    mod_height = (256 * 2) + 2

    module_pos = [  #
        (x_start - mod_width, y_start - mod_height - (i * (mod_height + 33)))
        for i in range(4)
    ] + [
        (-x_start, -y_start + (i * (mod_height + 33))) for i in range(4)
    ]
    orientations = [(-1, -1) for _ in range(4)] + [(1, 1) for _ in range(4)]

    return JUNGFRAUGeometry.from_module_positions(module_pos, orientations=orientations)


def test_write_read_crystfel_file(tmpdir):
    geom = jf4m_geometry()
    path = str(tmpdir / 'test.geom')
    geom.write_crystfel_geom(filename=path, photon_energy=9000,
                             adu_per_ev=0.0042, clen=0.101)

    loaded = JUNGFRAUGeometry.from_crystfel_geom(path)
    assert_geom_close(loaded, geom)

    # Load the geometry file with cfelpyutils and test the rigid groups
    geom_dict = load_crystfel_geometry(path)
    assert geom_dict['panels']['p0a0']['res'] == 1 / 75e-6
    p3a7 = geom_dict['panels']['p3a7']
    assert p3a7['min_ss'] == 256
    assert p3a7['max_ss'] == 511
    assert p3a7['min_fs'] == 768
    assert p3a7['max_fs'] == 1023


def test_get_pixel_positions():
    geom = jf4m_geometry()

    pixelpos = geom.get_pixel_positions()
    assert pixelpos.shape == (8, 512, 1024, 3)
    px = pixelpos[..., 0]
    py = pixelpos[..., 1]

    assert -0.09 < px.min() < -0.07
    assert 0.09 > px.max() > 0.07
    assert -0.09 < py.min() < -0.07
    assert 0.09 > py.max() > 0.07



from extra_geom import GenericGeometry

import extra_geom
import matplotlib.pyplot as plt
import numpy as np
import pytest

pixel_size = 0.01

simple_config = {'pixel_size': pixel_size,
                 'slow_pixels': 64,
                 'fast_pixels': 128,
                 'corner_coordinates': [pixel_size * np.array([5, 1, 0]),
                                        pixel_size * np.array([-133, 5, 0]),
                                        pixel_size * np.array([-63, -130, 0])],
                 'n_tiles_per_module': 2,
                 'ss_vec': np.array([1, 0, 0]),
                 'fs_vec': np.array([0, 1, 0]),
                 'tile_gap': 2 * pixel_size,
                 'tile_vec': [0, -1, 0]
                 }


@pytest.fixture
def simple():
    return GenericGeometry.from_simple_description(**simple_config)


def test_simple_geometry(simple):
    """ Tests that an object was created """
    assert isinstance(simple, extra_geom.detectors.GenericGeometry)
    assert simple.expected_data_shape == (3, 64, 256)


def test__tile_silce(simple):
    _slice = simple._tile_slice(1)
    assert isinstance(_slice, tuple)
    assert isinstance(_slice[0], slice)
    assert isinstance(_slice[1], slice)
    assert _slice == (slice(0, 64), slice(128, 256))


def test_write_crystfel_geom(simple, tmp_path):
    simple.write_crystfel_geom(tmp_path / 'test.geom')


def test_compare(simple):
    ax = simple.compare(simple)
    assert isinstance(ax, plt.Axes)


def test_inspect(simple):
    assert isinstance(simple.inspect(), plt.Axes)


def test_get_pixel_positions(simple):
    out = simple.get_pixel_positions()
    assert isinstance(out, np.ndarray)
    assert out.shape == simple.expected_data_shape + (3,)


def test_split_tiles(simple):
    data = np.zeros(simple.expected_data_shape)
    split_tile = simple.split_tiles(data)
    assert isinstance(split_tile, list)
    assert len(split_tile) == simple.n_tiles_per_module
    s1, s2, s3 = split_tile[0].shape
    assert (s1, s2 * simple.ss_tiles, s3 * simple.fs_tiles) == simple.expected_data_shape


def test_module_coords_to_tile(simple):
    """ expected data shape is (3, 64 [slow], 256 [fast: 2 tiles * 128 pixels])
    The points are:
    (5, 20),    <- t=0
    (10, 50),   <- t=0
    (30, 130),  <- t=1
    (60, 200),  <- t=1
    """
    slow_scan = np.array([5, 10, 30, 60])
    fast_scan = np.array([10, 50, 130, 200])
    tileno, tile_ss, tile_fs = simple._module_coords_to_tile(slow_scan, fast_scan)

    np.testing.assert_array_equal(tileno, [0, 0, 1, 1])
    np.testing.assert_array_equal(tile_ss, slow_scan)
    np.testing.assert_allclose(tile_fs, np.mod(fast_scan, 128))



import h5py
from matplotlib.axes import Axes
import numpy as np
from os.path import abspath, dirname, join as pjoin
from testpath import assert_isfile

from extra_geom import DSSC_1MGeometry
from .utils import assert_geom_close

tests_dir = dirname(abspath(__file__))
sample_xfel_geom = pjoin(tests_dir, 'dssc_geo_june19.h5')

# Made up numbers!
QUAD_POS = [
    (-130, 5),
    (-130, -125),
    (5, -125),
    (5, 5),
]


def test_inspect():
    geom = DSSC_1MGeometry.from_h5_file_and_quad_positions(
        sample_xfel_geom, QUAD_POS
    )
    # Smoketest
    ax = geom.inspect()
    assert isinstance(ax, Axes)


def test_snap_assemble_data():
    geom = DSSC_1MGeometry.from_h5_file_and_quad_positions(
        sample_xfel_geom, QUAD_POS
    )

    stacked_data = np.zeros((16, 128, 512))
    img, centre = geom.position_modules_fast(stacked_data)
    assert img.shape == (1281, 1099)
    assert tuple(centre) == (656, 552)
    assert np.isnan(img[0, 0])
    assert img[50, 50] == 0

def test_to_distortion_array():
    geom = DSSC_1MGeometry.from_h5_file_and_quad_positions(
        sample_xfel_geom, QUAD_POS
    )
    # Smoketest
    distortion = geom.to_distortion_array()
    assert isinstance(distortion, np.ndarray)
    assert distortion.shape == (2048, 512, 6, 3)

    # Coordinates in m, origin at corner; max x & y should be ~ 25cm
    assert 0.20 < distortion[..., 1].max() < 0.35
    assert 0.20 < distortion[..., 2].max() < 0.35
    assert 0.0 <= distortion[..., 1].min() < 0.01
    assert 0.0 <= distortion[..., 2].min() < 0.01

def test_get_pixel_positions():
    geom = DSSC_1MGeometry.from_h5_file_and_quad_positions(
        sample_xfel_geom, QUAD_POS
    )

    pixelpos = geom.get_pixel_positions()
    assert pixelpos.shape == (16, 128, 512, 3)

    px = pixelpos[..., 0]
    py = pixelpos[..., 1]

    assert -0.15 < px.min() < -0.1
    assert  0.15 > px.max() >  0.1
    assert -0.2 < py.min() < -0.12
    assert  0.2 > py.max() >  0.12

    # Odd-numbered rows in Q1 & Q2 should have 0.5 pixel higher x than the even.
    np.testing.assert_allclose(px[0, 1::2, 0] - px[0, 0::2, 0], 236e-6/2)


def test_read_write_xfel_file(tmpdir):
    geom = DSSC_1MGeometry.from_h5_file_and_quad_positions(
        sample_xfel_geom, QUAD_POS
    )
    path = str(tmpdir / 'dssc_geom.h5')
    quad_pos_out = geom.to_h5_file_and_quad_positions(path)

    assert_isfile(path)

    loaded = DSSC_1MGeometry.from_h5_file_and_quad_positions(path, quad_pos_out)
    assert_geom_close(loaded, geom)


def test_quad_positions_with_file():
    geom = DSSC_1MGeometry.from_h5_file_and_quad_positions(
        sample_xfel_geom, QUAD_POS
    )
    quad_pos_out = geom.quad_positions(sample_xfel_geom)

    np.testing.assert_allclose(quad_pos_out, QUAD_POS)


def test_quad_positions_no_file():
    geom = DSSC_1MGeometry.from_h5_file_and_quad_positions(
        sample_xfel_geom, QUAD_POS
    )
    # Smoketest - the results without passing a file in aren't usable yet
    quad_pos_out = geom.quad_positions()

    assert quad_pos_out.shape == (4, 2)



from cfelpyutils.crystfel_utils import load_crystfel_geometry
import h5py
from matplotlib.axes import Axes
import numpy as np
from os.path import abspath, dirname, join as pjoin
import pytest
from testpath import assert_isfile

from extra_geom import LPD_1MGeometry
from extra_geom.detectors import invert_xfel_lpd_geom
from .utils import assert_geom_close

tests_dir = dirname(abspath(__file__))


def test_write_read_crystfel_file(tmpdir):
    geom = LPD_1MGeometry.from_quad_positions(
        [(11.4, 299), (-11.5, 8), (254.5, -16), (278.5, 275)]
    )
    path = str(tmpdir / 'test.geom')
    geom.write_crystfel_geom(filename=path)

    with open(path, 'r') as f:
        contents = f.read()
    with open(path, 'w') as f:
        f.write('clen = 0.119\n')
        f.write('adu_per_eV = 0.0075\n')

        f.write(contents)
    # Load the geometry file with cfelpyutils and test the ridget groups
    loaded = LPD_1MGeometry.from_crystfel_geom(path)
    np.testing.assert_allclose(
        loaded.modules[0][0].corner_pos, geom.modules[0][0].corner_pos
    )
    np.testing.assert_allclose(loaded.modules[0][0].fs_vec, geom.modules[0][0].fs_vec)


    geom_dict = load_crystfel_geometry(path)
    quad_gr0 = ['p0a0', 'p0a1', 'p0a2', 'p0a3', 'p0a4', 'p0a5', 'p0a6', 'p0a7',
                'p0a8', 'p0a9', 'p0a10','p0a11', 'p0a12', 'p0a13', 'p0a14',
                'p0a15', 'p1a0', 'p1a1','p1a2', 'p1a3','p1a4','p1a5','p1a6',
                'p1a7', 'p1a8', 'p1a9', 'p1a10', 'p1a11', 'p1a12', 'p1a13',
                'p1a14', 'p1a15', 'p2a0', 'p2a1', 'p2a2', 'p2a3', 'p2a4', 'p2a5',
                'p2a6', 'p2a7', 'p2a8', 'p2a9', 'p2a10', 'p2a11', 'p2a12', 'p2a13',
                'p2a14','p2a15', 'p3a0', 'p3a1','p3a2', 'p3a3', 'p3a4', 'p3a5',
                'p3a6', 'p3a7', 'p3a8','p3a9', 'p3a10', 'p3a11', 'p3a12', 'p3a13',
                'p3a14', 'p3a15']
    assert geom_dict['rigid_groups']['p0'] == quad_gr0[:16]
    assert geom_dict['rigid_groups']['p3'] == quad_gr0[-16:]
    assert geom_dict['rigid_groups']['q0'] == quad_gr0


def test_read_write_xfel_file(tmpdir):
    quad_pos = [(11.4, 299), (-11.5, 8), (254.5, -16), (278.5, 275)]
    geom = LPD_1MGeometry.from_quad_positions(quad_pos)
    path = str(tmpdir / 'lpd_geom.h5')
    quad_pos_out = geom.to_h5_file_and_quad_positions(path)

    np.testing.assert_allclose(quad_pos_out, quad_pos)
    assert_isfile(path)

    loaded = LPD_1MGeometry.from_h5_file_and_quad_positions(path, quad_pos_out)
    assert_geom_close(loaded, geom)


def test_quad_positions_with_file():
    path = pjoin(tests_dir, 'lpd_mar_18.h5')
    quad_pos = [(11.4, 299), (-11.5, 8), (254.5, -16), (278.5, 275)]
    geom = LPD_1MGeometry.from_h5_file_and_quad_positions(path, quad_pos)

    np.testing.assert_allclose(geom.quad_positions(path), quad_pos)


def test_quad_positions_no_file():
    quad_pos = [(11.4, 299), (-11.5, 8), (254.5, -16), (278.5, 275)]
    geom = LPD_1MGeometry.from_quad_positions(quad_pos)

    np.testing.assert_allclose(geom.quad_positions(), quad_pos)


def test_offset():
    quad_pos = [(11.4, 299), (-11.5, 8), (254.5, -16), (278.5, 275)]
    geom = LPD_1MGeometry.from_quad_positions(quad_pos)
    y_orig = np.array([m[0].corner_pos[1] for m in geom.modules])

    # Uniform shift for all modules, all tiles
    all_shifted = geom.offset((0, 1e-3))
    y1 = np.array([m[0].corner_pos[1] for m in all_shifted.modules])
    np.testing.assert_allclose(y1, y_orig + 1e-3)

    # Select some modules
    q4_shifted = geom.offset((0, 2e-3), modules=np.s_[12:])
    y2 = np.array([m[0].corner_pos[1] for m in q4_shifted.modules])
    np.testing.assert_allclose(y2[:12], y_orig[:12])
    np.testing.assert_allclose(y2[12:], y_orig[12:] + 2e-3)

    quad_pos_modified = q4_shifted.quad_positions()
    np.testing.assert_allclose(quad_pos_modified[:3], quad_pos[:3])
    np.testing.assert_allclose(
        quad_pos_modified[3], np.array(quad_pos[3]) + [0, 2]  # quad positions in mm
    )

    # Per-module shift
    q3_shifted = geom.offset(np.repeat([
        (0, 0), (0, 0), (0, 2e-3), (0, 0),
    ], repeats=4, axis=0))
    y3 = np.array([m[0].corner_pos[1] for m in q3_shifted.modules])
    np.testing.assert_allclose(y3[:8], y_orig[:8])
    np.testing.assert_allclose(y3[8:12], y_orig[8:12] + 2e-3)
    np.testing.assert_allclose(y3[12:], y_orig[12:])

    # Per-tile shift
    shift = np.zeros((4, 16, 3), dtype=np.float64)
    shift[:, 5, 1] = 3e-3  # Shift T6 of each module in y dimension
    q2_t2_shifted = geom.offset(shift, modules=np.s_[4:8])
    y_t1 = np.array([m[0].corner_pos[1] for m in q2_t2_shifted.modules])
    np.testing.assert_allclose(y_t1, y_orig)
    y_t6 = np.array([m[5].corner_pos[1] for m in q2_t2_shifted.modules])
    y_t6_orig = np.array([m[5].corner_pos[1] for m in geom.modules])
    np.testing.assert_allclose(y_t6[:4], y_t6_orig[:4])
    np.testing.assert_allclose(y_t6[4:8], y_t6_orig[4:8] + 3e-3)
    np.testing.assert_allclose(y_t6[8:], y_t6_orig[8:])

    # Wrong number of modules
    with pytest.raises(ValueError):
        geom.offset(np.zeros((15, 2)))

    # Offsets for 16 modules, but only 4 selected
    with pytest.raises(ValueError):
        geom.offset(np.zeros((16, 16, 3)), modules=np.s_[:4])

    # Coordinates must be 2D or 3D
    with pytest.raises(ValueError):
        geom.offset(np.zeros(4))


def test_inspect():
    geom = LPD_1MGeometry.from_quad_positions(
        [(11.4, 299), (-11.5, 8), (254.5, -16), (278.5, 275)]
    )
    # Smoketest
    ax = geom.inspect()
    assert isinstance(ax, Axes)


def test_snap_assemble_data():
    geom = LPD_1MGeometry.from_quad_positions(
        [(11.4, 299), (-11.5, 8), (254.5, -16), (278.5, 275)]
    )

    stacked_data = np.zeros((16, 256, 256))
    img, centre = geom.position_modules_fast(stacked_data)
    assert img.shape == (1202, 1104)
    assert tuple(centre) == (604, 547)
    assert np.isnan(img[0, 0])
    assert img[50, 50] == 0

def test_to_distortion_array():
    geom = LPD_1MGeometry.from_quad_positions(
        [(11.4, 299), (-11.5, 8), (254.5, -16), (278.5, 275)]
    )
    # Smoketest
    distortion = geom.to_distortion_array()
    assert isinstance(distortion, np.ndarray)
    assert distortion.shape == (4096, 256, 4, 3)

    # Coordinates in m, origin at corner; max x & y should be ~ 50cm
    assert 0.40 < distortion[..., 1].max() < 0.70
    assert 0.40 < distortion[..., 2].max() < 0.70
    assert 0.0 <= distortion[..., 1].min() < 0.01
    assert 0.0 <= distortion[..., 2].min() < 0.01

def test_data_coords_to_positions():
    geom = LPD_1MGeometry.from_quad_positions(
        [(11.4, 299), (-11.5, 8), (254.5, -16), (278.5, 275)]
    )

    module_no = np.zeros(16, dtype=np.int16)
    # Points near the centre of each tile
    slow_scan = np.tile(np.linspace(16, 240, num=8, dtype=np.float32), 2)
    fast_scan = np.array([64, 192], dtype=np.float32).repeat(8)

    tileno, tile_ss, tile_fs = geom._module_coords_to_tile(slow_scan, fast_scan)
    np.testing.assert_allclose(tileno,
                       [7, 6, 5, 4, 3, 2, 1, 0, 8, 9, 10, 11, 12, 13, 14, 15])
    np.testing.assert_allclose(tile_ss, 16)
    np.testing.assert_allclose(tile_fs, 64)

    res = geom.data_coords_to_positions(module_no, slow_scan, fast_scan)

    assert res.shape == (16, 3)

    resx, resy, resz = res.T

    np.testing.assert_allclose(resz, 0)

    assert (np.diff(resy[:8]) > 0).all()  # T1-T8 Monotonically increasing
    assert (np.diff(resy[8:]) > 0).all()  # T9-T16 Monotonically increasing
    assert -0.128 > resx.max() > resx.min() > -0.280

def test_invert_xfel_lpd_geom(tmpdir):
    src_file = pjoin(tests_dir, 'lpd_mar_18.h5')
    dst_file = pjoin(str(tmpdir), 'lpd_inverted.h5')
    invert_xfel_lpd_geom(src_file, dst_file)
    with h5py.File(src_file, 'r') as fsrc, h5py.File(dst_file, 'r') as fdst:
        np.testing.assert_array_equal(
            fsrc['Q1/M1/Position'][:], -1 * fdst['Q1/M1/Position'][:]
        )
        np.testing.assert_array_equal(
            fsrc['Q1/M1/T07/Position'][:], -1 * fdst['Q1/M1/T07/Position'][:]
        )



"""DEPRECATED: LPD geometry. Use extra_geom.LPD_1MGeometry instead.
"""
from copy import copy
import h5py
from itertools import product
import numpy as np
import sys
from textwrap import indent


def splitChannelDataIntoTiles(channelData, clockwiseOrder=False):
    """Splits the raw channel data into indiviual tiles

    Args
    ----

    channelData : ndarray
        Raw channel data. Must have shape (256, 256)

    clockwiseOrder : bool, optional
        If set to True, the sequence of tiles is given
        in the clockwise order starting with the top
        right tile (LPD standard). If set to false, tile
        data is returned in reading order

    Returns
    -------

    ndarray
        Same data, but reshaped into (12, 32, 128)
    """
    extra_dims = channelData.shape[:-2]
    a = np.asarray(np.split(channelData, 8, axis=-2))
    a = np.asarray(np.split(a, 2, axis=-1))
    a = np.reshape(a, (16,) + extra_dims + (32, 128))
    orderedTiles = a

    if clockwiseOrder:
        # The official LPD tile order is clockwise from the top right tile.
        # We need them in this order to apply the right offset to the right tile.
        readingOrderToClockwise = list(range(7, -1, -1)) + list(range(8, 16))
        orderedTiles = orderedTiles[readingOrderToClockwise]
    return orderedTiles


class GeometryFragment:
    def __init__(self, offset, children):
        self.offset = offset  # in m
        self.children = children

    def _str_lines(self):
        r = []
        for name, child in sorted(self.children.items()):
            r.append("{}: {}".format(name, child.offset))
            r.extend(indent(str(child), '  ').splitlines())
        return r

    def __str__(self):
        return '\n'.join(self._str_lines())

    def find_offset(self, name_parts):
        if name_parts:
            child = self.children[name_parts[0]]
            return self.offset + child.find_offset(name_parts[1:])
        return self.offset

    def iter_leaf_paths(self):
        if not self.children:
            yield ()
            return
        for k, v in sorted(self.children.items()):
            for subpath in v.iter_leaf_paths():
                yield (k,) + subpath

    @classmethod
    def from_h5_group(cls, group, unit=1e-3):
        children = {
            key: GeometryFragment.from_h5_group(val)
            for (key, val) in group.items()
            if isinstance(val, h5py.Group)
        }
        return cls(group['Position'][:] * unit, children)


class LPDGeometry(GeometryFragment):
    pixel_size = 0.5e-3  # Meter: 0.5 Millimeter

    @classmethod
    def from_h5_file_and_quad_positions(cls, file, positions, unit=1e-3):
        children = {}
        for n, position in enumerate(positions, start=1):
            quad = 'Q%d' % n
            modules = {
                name: GeometryFragment.from_h5_group(group, unit=unit)
                for (name, group) in file[quad].items()
            }
            children[quad] = GeometryFragment(np.asarray(position) * unit, modules)

        return cls(np.asarray((0, 0)), children)

    def position_all_modules(self, data):
        """Assemble data from this detector according to where the pixels are.

        Parameters
        ----------

        data : ndarray
          The last three dimensions should be channelno, pixel_y, pixel_x
          (lengths 16, 256, 256).
          Other dimensions before these will be preserved in the output.

        Returns
        -------
        out : ndarray
          Array with the one dimension fewer than the input.
          The last two dimensions represent pixel y and x in the detector space.
        centre : ndarray
          (x, y) pixel location of the detector centre in this geometry.
        """
        assert data.shape[-3:] == (16, 256, 256)
        size_xy, centre = self._plotting_dimensions()
        size_yx = size_xy[::-1]
        out = np.empty(data.shape[:-3] + size_yx, dtype=data.dtype)
        out[:] = np.nan

        for channelno in range(16):
            module_data = data[..., channelno, :, :]

            Q = 'Q{:d}'.format(channelno // 4 + 1)
            M = 'M{:d}'.format(channelno % 4 + 1)
            module_tiles_data = splitChannelDataIntoTiles(
                module_data, clockwiseOrder=True
            )
            for tileno, tile_data in enumerate(module_tiles_data, start=1):
                T = "T{:02d}".format(tileno)
                position = self.find_offset((Q, M, T))
                # Convert physical position (m) to pixel location:
                x0, y0 = (position // self.pixel_size) + centre
                x0, y0 = int(x0), int(y0)

                out[..., y0:y0 + tile_data.shape[-2],
                         x0:x0 + tile_data.shape[-1]] = tile_data[..., ::-1, ::-1]

        return out, centre

    def _plotting_dimensions(self):
        """Calculate appropriate dimensions for plotting assembled data

        Returns (size_x, size_y), (centre_x, centre_y)
        """
        min_x = max_x = min_y = max_y = 0
        for path in self.iter_leaf_paths():
            x, y = self.find_offset(path)
            min_x = min(min_x, x)
            max_x = max(max_x, x)
            min_y = min(min_y, y)
            max_y = max(max_y, y)

        # Convert physical distances to pixels
        # Add 20px margin, plus the 128*32 tile size
        min_x = int(min_x // self.pixel_size) - 20
        max_x = int(max_x // self.pixel_size) + 128 + 20
        min_y = int(min_y // self.pixel_size) - 20
        max_y = int(max_y // self.pixel_size) + 32 + 20

        size = (max_x - min_x), (max_y - min_y)
        centre = np.asarray([-min_x, -min_y])
        return size, centre

    def plot_data(self, modules_data):
        """Plot data from the detector using this geometry.

        Returns a matplotlib figure.

        Parameters
        ----------

        modules_data : ndarray
          Should have exactly 3 dimensions: channelno, pixel_y, pixel_x
          (lengths 16, 256, 256).
        """
        from matplotlib.cm import viridis
        from matplotlib.backends.backend_agg import FigureCanvasAgg
        from matplotlib.figure import Figure

        fig = Figure((10, 10))
        FigureCanvasAgg(fig)
        ax = fig.add_subplot(1, 1, 1)
        my_viridis = copy(viridis)
        # Use a dark grey for missing data
        my_viridis.set_bad('0.25', 1.0)

        res, centre = self.position_all_modules(modules_data)
        ax.imshow(res, cmap=my_viridis)

        cx, cy = centre
        ax.hlines(cy, cx - 20, cx + 20, colors='w', linewidths=1)
        ax.vlines(cx, cy - 20, cy + 20, colors='w', linewidths=1)
        return fig

    def inspect(self):
        """Plot the 2D layout of this detector geometry.

        Returns a matplotlib Figure object.
        """
        from matplotlib.backends.backend_agg import FigureCanvasAgg
        from matplotlib.figure import Figure

        fig = Figure((10, 10))
        FigureCanvasAgg(fig)
        ax = fig.add_subplot(1, 1, 1)

        size_xy, centre = self._plotting_dimensions()
        size_yx = size_xy[::-1]  # Our array indexing is [y, x]
        bg = np.ones(size_yx + (3,), dtype='f4')

        # Mark the centre location
        cx, cy = centre
        ax.hlines(cy, cx - 100, cx + 100, colors='0.75', linewidths=2)
        ax.vlines(cx, cy - 100, cy + 100, colors='0.75', linewidths=2)

        # Show where detector elements (modules & tiles) are positioned.
        for (Q, M) in product(range(1, 5), range(1, 5)):
            position = self.find_offset(('Q%d' % Q, 'M%d' % M))
            xm, ym = (position // self.pixel_size) + centre
            ax.text(xm, ym, 'Q%dM%d' % (Q, M), color='k', ha='left', va='top',
                    bbox={'facecolor': 'w'})
            for T in range(1, 17):
                position = self.find_offset(('Q%d' % Q, 'M%d' % M, 'T%02d' % T))
                xt, yt = (position // self.pixel_size) + centre
                xt, yt = int(xt), int(yt)
                # Draw a light green block for each tile
                bg[yt : yt + 32, xt : xt + 128] = (0.75, 1.0, 0.75)
                # Label specific tiles to show the ordering
                if T in [1, 8, 9]:
                    ax.text(xt, yt, str(T), va='top', ha='left')

        ax.imshow(bg, origin='upper')
        ax.set_title('LPD detector geometry')
        return fig


if __name__ == '__main__':
    quadpos = [(-11.4, -299), (11.5, -8), (-254.5, 16), (-278.5, -275)]  # MAR 18
    with h5py.File(sys.argv[1], 'r') as f:
        geom = LPDGeometry.from_h5_file_and_quad_positions(f, quadpos, unit=1e-3)

    print(geom)
    print('Q2/M1/T07:', geom.find_offset(('Q2', 'M1', 'T07')))



"""Write geometry in CrystFEL format.
"""
from itertools import product
import numpy as np

HEADER_TEMPLATE = """\
; {detector} geometry file written by EXtra-geom {version}
; You may need to edit this file to add:
; - data and mask locations in the file
; - mask_good & mask_bad values to interpret the mask
; - adu_per_eV & photon_energy
; - clen (detector distance)
;
; See: http://www.desy.de/~twhite/crystfel/manual-crystfel_geometry.html

{paths}
{frame_dim}
res = {resolution} ; pixels per metre

; Beam energy in eV
{photon_energy}

; Camera length, aka detector distance
{clen}

; Analogue Digital Units per eV
{adu_per_ev}
"""


PANEL_TEMPLATE = """
{dims}
{name}/min_fs = {min_fs}
{name}/min_ss = {min_ss}
{name}/max_fs = {max_fs}
{name}/max_ss = {max_ss}
{name}/fs = {fs_vec}
{name}/ss = {ss_vec}
{name}/corner_x = {corner_x}
{name}/corner_y = {corner_y}
{name}/coffset = {coffset}
"""


def _crystfel_format_vec(vec):
    """Convert an array of 3 numbers to CrystFEL format like "+1.0x -0.1y"
    """
    s = '{:+}x {:+}y'.format(*vec[:2])
    if vec[2] != 0:
        s += ' {:+}z'.format(vec[2])
    return s


def frag_to_crystfel(fragment, p, a, ss_slice, fs_slice, dims, pixel_size):
    tile_name = 'p{}a{}'.format(p, a)
    c = fragment.corner_pos / pixel_size
    dim_list = []
    for num, value in dims.items():
        if value == 'modno':
            key = p
        else:
            key = value
        dim_list.append('{}/dim{} = {}'.format(tile_name, num, key))

    return PANEL_TEMPLATE.format(
        dims='\n'.join(dim_list),
        name=tile_name,
        min_ss=ss_slice.start,
        max_ss=ss_slice.stop - 1,
        min_fs=fs_slice.start,
        max_fs=fs_slice.stop - 1,
        ss_vec=_crystfel_format_vec(fragment.ss_vec / pixel_size),
        fs_vec=_crystfel_format_vec(fragment.fs_vec/ pixel_size),
        corner_x=c[0],
        corner_y=c[1],
        coffset=c[2],
    )

def write_crystfel_geom(self, filename, *,
                        data_path='/entry_1/instrument_1/detector_1/data',
                        mask_path=None, dims=('frame', 'modno', 'ss', 'fs'),
                        nquads=4, adu_per_ev=None, clen=None,
                        photon_energy=None):
    """Write this geometry to a CrystFEL format (.geom) geometry file.
    """
    from . import __version__

    if adu_per_ev is None:
        adu_per_ev_str = '; adu_per_eV = SET ME'
        # TODO: adu_per_ev should be fixed for each detector, we should
        #       find out the values and set them.
    else:
        adu_per_ev_str = 'adu_per_eV = {}'.format(adu_per_ev)

    if clen is None:
        clen_str = '; clen = SET ME'
    else:
        clen_str = 'clen = {}'.format(clen)

    if photon_energy is None:
        photon_energy_str = '; photon_energy = SET ME'
    else:
        photon_energy_str = 'photon_energy = {}'.format(photon_energy)

    # Get the frame dimension
    tile_dims = {}

    frame_dim = None
    for nn, dim_name in enumerate(dims):
        if dim_name == 'frame':
            frame_dim = 'dim{} = %'.format(nn)
        else:
            tile_dims[nn] = dim_name
    if frame_dim is None:
        raise ValueError('No frame dimension given')

    panel_chunks = []
    for p, module in enumerate(self.modules):
        for a, fragment in enumerate(module):
            ss_slice, fs_slice = self._tile_slice(a)
            if 'modno' not in dims:
                # If we don't have a modno dimension, assume modules are
                # concatenated along the slow-scan dim, e.g. AGIPD (8192, 128)
                module_offset = p * self.expected_data_shape[1]
                ss_slice = slice(
                    ss_slice.start + module_offset,
                    ss_slice.stop + module_offset
                )

            panel_chunks.append(frag_to_crystfel(
                fragment, p, a, ss_slice, fs_slice, tile_dims, self.pixel_size
            ))

    resolution = 1.0 / self.pixel_size  # Pixels per metre

    paths = dict(data=data_path)
    if mask_path:
        paths['mask'] = mask_path
    path_str = '\n'.join('{} = {} ;'.format(i, j) for i, j in paths.items())

    with open(filename, 'w') as f:
        f.write(HEADER_TEMPLATE.format(
            detector=self.detector_type_name,
            version=__version__,
            paths=path_str,
            frame_dim=frame_dim,
            resolution=resolution,
            adu_per_ev=adu_per_ev_str,
            clen=clen_str,
            photon_energy=photon_energy_str
        ))
        rigid_groups = get_rigid_groups(self, nquads=nquads)
        f.write(rigid_groups)
        for chunk in panel_chunks:
            f.write(chunk)

def get_rigid_groups(geom, nquads=4):
    """Create string for rigid groups definition."""

    lines = []

    module_groups = {
        f'p{p}': [f"p{p}a{a}" for a in range(len(tiles))]
        for p, tiles in enumerate(geom.modules)
    }
    lines += [
        f'rigid_group_{k} = ' + ','.join(v) for k, v in module_groups.items()
    ] + [
        'rigid_group_collection_modules = ' + ','.join(module_groups),
        ''
    ]

    if nquads:
        if len(geom.modules) % nquads:
            raise ValueError(
                f"Can't divide {len(geom.modules)} modules into {nquads} groups"
            )
        prod = product(range(len(geom.modules)), range(geom.n_tiles_per_module))
        all_panels = [f"p{p}a{a}" for (p, a) in prod]
        quad_groups = {
            f'q{q}': panels
            for q, panels in enumerate(np.array_split(all_panels, nquads))
        }
        lines += [
            f'rigid_group_{k} = ' + ','.join(v) for k, v in quad_groups.items()
        ] + [
            'rigid_group_collection_quads = ' + ','.join(quad_groups),
            ''
        ]

    return '\n'.join(lines)



import numpy as np

def all_tiles(geom):
    for mod in geom.modules:
        yield from mod

def tile_positions(geom):
    return np.stack([t.corner_pos for t in all_tiles(geom)])

def tile_vectors(geom):
    return np.stack([(t.ss_vec, t.fs_vec) for t in all_tiles(geom)])

def assert_geom_close(g1, g2):
    assert len(g1.modules) == len(g2.modules)
    assert {len(m) for m in g1.modules} == {len(m) for m in g2.modules}

    np.testing.assert_allclose(tile_positions(g1), tile_positions(g2))
    np.testing.assert_allclose(tile_vectors(g1), tile_vectors(g2))



"""Tests for deprecated extra_geom.geometry module."""

import h5py
from matplotlib.figure import Figure
import numpy as np
from os.path import abspath, dirname, join as pjoin

from extra_geom.lpd_old import LPDGeometry

tests_dir = dirname(abspath(__file__))

# These coordinates are backwards - x increasing beam-right, y increasing
# downwards. Use LPD_1MGeometry instead, which expects coordinates
# in the standard axis directions.
quadpos = [(-11.4, -299), (11.5, -8), (-254.5, 16), (-278.5, -275)]


def test_inspect():
    with h5py.File(pjoin(tests_dir, 'lpd_mar_18.h5'), 'r') as f:
        geom = LPDGeometry.from_h5_file_and_quad_positions(f, quadpos)

    # Smoketest
    fig = geom.inspect()
    assert isinstance(fig, Figure)


def test_position_all_modules():
    with h5py.File(pjoin(tests_dir, 'lpd_mar_18.h5'), 'r') as f:
        geom = LPDGeometry.from_h5_file_and_quad_positions(f, quadpos)

    stacked_data = np.zeros((16, 256, 256))
    img, centre = geom.position_all_modules(stacked_data)
    assert 1000 < img.shape[0] < 2000
    assert 1000 < img.shape[1] < 2000
    assert np.isnan(img[0, 0])
    assert img[200, 200] == 0



"""This deals with geometry approximated to fit a regular 2D pixel grid

This is useful to copy data from multiple modules into a single image array.
This module is not a public API: it's used internally by the classes in
extra_geom.detectors.
"""
from copy import copy
from itertools import chain
import numpy as np
import sys


class GridGeometryFragment:
    """Holds the 2D axis-aligned position and orientation of one detector tile.

    This is used in 'snapped' geometry which efficiently assembles a detector
    image into a 2D array.

    These coordinates are all (y, x), suitable for indexing a numpy array.

    ss_vec and fs_vec must be length 1 vectors in either positive or negative
    x or y direction. In the output array, the fast scan dimension is always x.
    So if the input data is oriented with fast-scan vertical, we need to
    transpose it first.

    Regardless of transposition, we may also need to flip the data on one or
    both axes; the fs_order and ss_order variables handle this.
    """
    def __init__(self, corner_pos, ss_vec, fs_vec, ss_pixels, fs_pixels):
        self.ss_vec = ss_vec
        self.fs_vec = fs_vec
        self.ss_pixels = ss_pixels
        self.fs_pixels = fs_pixels

        if fs_vec[0] == 0:
            # Fast scan is x dimension: Flip without transposing
            fs_order = fs_vec[1]
            ss_order = ss_vec[0]
            self.transform = lambda arr: arr[..., ::ss_order, ::fs_order]
            corner_shift = np.array([
                min(ss_order, 0) * self.ss_pixels,
                min(fs_order, 0) * self.fs_pixels
            ])
            self.pixel_dims = (self.ss_pixels, self.fs_pixels)
        else:
            # Fast scan is y : Transpose so fast scan -> x and then flip
            fs_order = fs_vec[0]
            ss_order = ss_vec[1]
            self.transform = lambda arr: arr.swapaxes(-1, -2)[..., ::fs_order, ::ss_order]
            corner_shift = np.array([
                min(fs_order, 0) * self.fs_pixels,
                min(ss_order, 0) * self.ss_pixels
            ])
            self.pixel_dims = (self.fs_pixels, self.ss_pixels)
        self.corner_idx = tuple(corner_pos + corner_shift)

    def offset(self, y_x) -> 'GridGeometryFragment':
        new = copy(self)
        new.corner_idx = tuple(np.array(self.corner_idx) + y_x)
        return new


class SnappedGeometry:
    """Detector geometry approximated to align modules to a 2D grid

    The coordinates used in this class are (y, x) suitable for indexing a
    Numpy array; this does not match the (x, y, z) coordinates in the more
    precise geometry above.
    """
    def __init__(self, modules, geom, centre):
        self.modules = modules
        self.geom = geom
        self.centre = centre

        # The fragments here are already shifted so corner_idx starts from 0 in
        # each dim, so the max outer edges define the output image size.
        self.size_yx = tuple(np.max([
            np.array(frag.corner_idx) + np.array(frag.pixel_dims)
            for frag in chain(*modules)
        ], axis=0))

    def make_output_array(self, extra_shape=(), dtype=np.float32):
        """Make an output array for self.position_modules()
        """
        shape = extra_shape + self.size_yx
        return np.full(shape, np.nan, dtype=dtype)

    def position_modules(self, data, out=None, threadpool=None):
        """Implementation for position_modules_fast
        """
        nmod = self.geom.expected_data_shape[0]
        if isinstance_no_import(data, 'xarray', 'DataArray'):
            # Input is an xarray labelled array
            modnos = data.coords.get('module')
            if modnos is None:
                raise ValueError(
                    "xarray arrays should have a dimension named 'module'"
                )
            modnos = modnos.values  # xarray -> numpy
            min_mod, max_mod = modnos.min(), modnos.max()
            if min_mod < 0 or max_mod >= nmod:
                raise ValueError(
                    f"module number labels should be in the range 0-{nmod-1} "
                    f"(found {min_mod}-{max_mod})"
                )
            assert data.shape[-2:] == self.geom.expected_data_shape[-2:]

            mod_dim_ix = data.dims.index('module')
            extra_shape = data.shape[:mod_dim_ix] + data.shape[mod_dim_ix+1:-2]
            get_mod_data = lambda i: data.sel(module=i).values
        else:
            # Input is a numpy array (or similar unlabelled array)
            assert data.shape[-3:] == self.geom.expected_data_shape
            modnos = range(nmod)
            extra_shape = data.shape[:-3]
            get_mod_data = lambda i: data[..., i, :, :]

        if out is None:
            out = self.make_output_array(extra_shape, data.dtype)
        else:
            assert out.shape == extra_shape + self.size_yx
            if not np.can_cast(data.dtype, out.dtype, casting='safe'):
                raise TypeError("{} cannot be safely cast to {}".
                                format(data.dtype, out.dtype))

        copy_pairs = []
        for modno in modnos:
            module = self.modules[modno]
            tiles_data = self.geom.split_tiles(get_mod_data(modno))
            for tile, tile_data in zip(module, tiles_data):
                y, x = tile.corner_idx
                h, w = tile.pixel_dims

                copy_pairs.append((
                    out[..., y : y + h, x : x + w], tile.transform(tile_data)
                ))

        if threadpool is not None:
            def copy_data(pair):
                dst, src = pair
                dst[:] = src
            # concurrent.futures map() is async, so call list() to wait for it
            list(threadpool.map(copy_data, copy_pairs))
        else:
            for dst, src in copy_pairs:
                dst[:] = src

        return out, self.centre

    def position_modules_symmetric(self, data, out=None, threadpool=None):
        """Assemble data so the centre is in the middle of the output array"""
        assert data.shape[-3:] == self.geom.expected_data_shape
        min_shape = np.stack([self.centre * 2, self.size_yx]).max(axis=0)
        if out is None:
            img_shape = min_shape
            out = np.full(data.shape[:-3] + tuple(img_shape), np.nan, dtype=data.dtype)
        else:
            assert out.shape[:-2] == data.shape[:-3]
            img_shape = np.array(out.shape[-2:])
            if (img_shape < np.array(min_shape)).any():
                raise ValueError(
                    f"Output shape {img_shape} less than required {min_shape}"
                )

        y, x = (img_shape // 2) - self.centre  # Find offset
        h, w = self.size_yx
        self.position_modules(
            data, out[..., y:y+h, x:x+w], threadpool=threadpool
        )
        return out

    def plot_data(self,
                  modules_data, *,
                  axis_units='px',
                  frontview=True,
                  ax=None,
                  figsize=None,
                  colorbar=False,
                  **kwargs):
        """Implementation for plot_data_fast
        """
        from matplotlib.cm import viridis
        import matplotlib.pyplot as plt

        if axis_units not in {'px', 'm'}:
            raise ValueError("axis_units must be 'px' or 'm', not {!r}"
                             .format(axis_units))

        res, centre = self.position_modules(modules_data)
        min_y, min_x = -centre
        max_y, max_x = np.array(res.shape) - centre

        _extent = np.array((min_x - 0.5, max_x + 0.5, min_y - 0.5, max_y + 0.5))
        cross_size = 20
        if axis_units == 'm':
            _extent[:2] *= self.geom._pixel_shape[0]  # x
            _extent[2:] *= self.geom._pixel_shape[1]  # y
            cross_size *= self.geom.pixel_size

        # Use a dark grey for missing data
        _cmap = copy(viridis)
        _cmap.set_bad('0.25', 1.0)

        kwargs.setdefault('cmap', _cmap)
        kwargs.setdefault('extent', _extent)
        kwargs.setdefault('origin', 'lower')

        if ax is None:
            fig = plt.figure(figsize=figsize or (10, 10))
            ax = fig.add_subplot(1, 1, 1)

        im = ax.imshow(res, **kwargs)
        if isinstance(colorbar, dict) or colorbar is True:
            if isinstance(colorbar, bool):
                colorbar = {}
            colorbar = plt.colorbar(im, ax=ax, **colorbar)

        ax.set_xlabel('metres' if axis_units == 'm' else 'pixels')
        ax.set_ylabel('metres' if axis_units == 'm' else 'pixels')

        if frontview:
            ax.invert_xaxis()

        # Draw a cross at the centre
        ax.hlines(0, -cross_size, +cross_size, colors='w', linewidths=1)
        ax.vlines(0, -cross_size, +cross_size, colors='w', linewidths=1)
        return ax


def isinstance_no_import(obj, mod: str, cls: str):
    """Check if isinstance(obj, mod.cls) without loading mod"""
    m = sys.modules.get(mod)
    if m is None:
        return False

    return isinstance(obj, getattr(m, cls))



#!/usr/bin/env python
import os.path as osp
import re
from setuptools import setup, find_packages
import sys


def get_script_path():
    return osp.dirname(osp.realpath(sys.argv[0]))


def read(*parts):
    return open(osp.join(get_script_path(), *parts)).read()


def find_version(*parts):
    vers_file = read(*parts)
    match = re.search(r'^__version__ = "(\d+\.\d+\.\d+)"', vers_file, re.M)
    if match is not None:
        return match.group(1)
    raise RuntimeError("Unable to find version string.")


setup(name="EXtra-geom",
      version=find_version("extra_geom", "__init__.py"),
      author="European XFEL GmbH",
      author_email="da-support@xfel.eu",
      maintainer="Thomas Michelat",
      url="https://github.com/European-XFEL/EXtra-geom",
      description="Tools to work with EuXFEL detector geometry and assemble detector images",
      long_description=read("README.md"),
      long_description_content_type='text/markdown',
      license="BSD-3-Clause",
      packages=find_packages(),
      package_data={
          'extra_geom.tests': ['dssc_geo_june19.h5', 'lpd_mar_18.h5'],
      },
      install_requires=[
          'cfelpyutils>=0.92',
          'h5py>=2.7.1',
          'matplotlib',
          'numpy',
          'scipy',
      ],
      extras_require={
          'docs': [
              'sphinx',
              'nbsphinx',
              'ipython',  # For nbsphinx syntax highlighting
              'sphinxcontrib_github_alt',
          ],
          'test': [
              'pytest',
              'pytest-cov',
              'coverage',
              'nbval',
              'testpath',
              'xarray',
              'EXtra-data',
          ]
      },
      python_requires='>=3.6',
      classifiers=[
          'Intended Audience :: Developers',
          'Intended Audience :: Science/Research',
          'License :: OSI Approved :: BSD License',
          'Programming Language :: Python :: 3',
          'Topic :: Scientific/Engineering :: Information Analysis',
          'Topic :: Scientific/Engineering :: Physics',
      ],
      project_urls={
          'Documentation': 'https://extra-geom.readthedocs.io/en/latest/',
          'Changelog': 'https://extra-geom.readthedocs.io/en/latest/changelog.html',
      },
)
